//------------------------------------------------------------------------------------------------------------------------------------------
// Module that handles controls and movement for the player.
//
// Note that for Final Doom the movement code was changed in many ways here versus the original PSX Doom, with movement amounts being
// expressed in terms of 1 vblank instead of 4 vblanks - presumably done as an optimization avoid many divisions. The movement speed and
// friction also behaves a little differently in the case of Final Doom. This module supports both the Doom and Final Doom way of doing
// things in order to achieve demo compatibility in both cases, hence the code is a little more complex because of this.
//------------------------------------------------------------------------------------------------------------------------------------------
#include "p_user.h"

#include "Doom/Base/i_main.h"
#include "Doom/Base/m_fixed.h"
#include "Doom/Base/s_sound.h"
#include "Doom/Base/sounds.h"
#include "Doom/d_main.h"
#include "Doom/Renderer/r_local.h"
#include "Doom/Renderer/r_main.h"
#include "Doom/UI/st_main.h"
#include "g_game.h"
#include "info.h"
#include "p_local.h"
#include "p_map.h"
#include "p_mobj.h"
#include "p_pspr.h"
#include "p_slide.h"
#include "p_spec.h"
#include "p_tick.h"
#include "PsyDoom/Config/Config.h"
#include "PsyDoom/Controls.h"
#include "PsyDoom/DemoPlayer.h"
#include "PsyDoom/Game.h"
#include "PsyDoom/Input.h"
#include "PsyDoom/PlayerPrefs.h"
#include "PsyDoom/PsxPadButtons.h"
#include "PsyDoom/Utils.h"

#include <algorithm>
#include <chrono>
#include <cmath>

// Accelerating turn speeds: normal (Doom, for 4 vblanks)
static constexpr fixed_t ANGLE_TURN_DOOM[10] = {
    300, 300, 500, 500, 600, 700, 800, 900, 900, 1000
};

// Accelerating turn speeds: normal (Final Doom, for 1 vblank)
static constexpr fixed_t ANGLE_TURN_FDOOM[10] = {
    75, 75, 125, 125, 150, 175, 200, 225, 225, 250
};

// Accelerating turn speeds: fast (Doom, for 4 vblanks)
static constexpr fixed_t FAST_ANGLE_TURN_DOOM[10] = {
    800, 800, 900, 1000, 1000, 1200, 1200, 1300, 1300, 1400
};

// Accelerating turn speeds: fast (Final Doom, for 1 vblank)
static constexpr fixed_t FAST_ANGLE_TURN_FDOOM[10] = {
    200, 200, 225, 250, 250, 300, 300, 325, 325, 350
};

static constexpr fixed_t STOPSPEED              = FRACUNIT / 16;                    // Stop the player completely if velocity falls under this amount
static constexpr fixed_t FRICTION               = 0xd200;                           // Friction amount to apply to player movement: approximately 0.82
static constexpr fixed_t FORWARD_MOVE_DOOM[2]   = { 0x40000, 0x60000 };             // Movement speeds: forward/back (Doom, for 4 vblanks)
static constexpr fixed_t FORWARD_MOVE_FDOOM[2]  = { 0x0E000, 0x16000 };             // Movement speeds: forward/back (Final Doom, for 1 vblank)
static constexpr fixed_t SIDE_MOVE_DOOM[2]      = { 0x38000, 0x58000 };             // Movement speeds (Doom, for 4 vblanks): strafe left/right
static constexpr fixed_t SIDE_MOVE_FDOOM[2]     = { 0x0E000, 0x16000 };             // Movement speeds (Final Doom, for 1 vblank): strafe left/right
static constexpr int32_t TURN_ACCEL_TICS        = C_ARRAY_SIZE(ANGLE_TURN_DOOM);    // Number of tics/stages in turn acceleration before it hits max speed
static constexpr fixed_t MAXBOB                 = 16 * FRACUNIT;                    // Maximum amount of view bobbing per frame (16 pixels)

// Flag set to true when the player is on the ground
static bool gbOnGround;

#if PSYDOOM_MODS
    // Convenience typedef
    typedef std::chrono::high_resolution_clock::time_point time_point_t;

    // How much turning done which hasn't been merged into the player's map object.
    // This eventually gets rolled into the player's actual angle, but is used in the intermediate to update the renderer.
    // It basically allows for framerate uncapped turning movement with very little latency, outside of the 30 Hz update loop.
    angle_t gPlayerUncommittedTurning;

    // Only used in network games: the view angle the player will use on the next frame.
    // Used in networked games to preserve any turning the user did, even though it won't be used until the next frame.
    // The view angle for rendering is allowed to be 1 frame ahead of where it actually is.
    angle_t gPlayerNextTickViewAngle;

    // When we last did framerate uncapped turning movements for the current player
    static time_point_t gLastPlayerTurnTime;
#endif

//------------------------------------------------------------------------------------------------------------------------------------------
// Attempts to move the given player map object according to it's current velocity.
// Also crosses special lines in the process, and interacts with special things touched.
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_PlayerMove(mobj_t& mobj) noexcept {
    // This is the amount to be moved
    fixed_t moveDx;
    fixed_t moveDy;

    if (Game::gSettings.bUseFinalDoomPlayerMovement) {
        moveDx = mobj.momx;
        moveDy = mobj.momy;
    } else {
        const int32_t elapsedVBlanks = gPlayersElapsedVBlanks[gPlayerNum];
        moveDx = d_rshift<2>(mobj.momx) * elapsedVBlanks;
        moveDy = d_rshift<2>(mobj.momy) * elapsedVBlanks;
    }

    // Do the sliding movement
    gpSlideThing = &mobj;
    P_SlideMove();

    // If we did not succeed in doing a sliding movement, or cannot go to the suggested location then try stairstepping.
    // With stairstepping we try to do the x and y axis movements alone, instead of together:
    const bool bSlideMoveBlocked = ((gSlideX == mobj.x) && (gSlideY == mobj.y));

    if (bSlideMoveBlocked || (!P_TryMove(mobj, gSlideX, gSlideY))) {
        // Clamp the movement amount so we don't go beyond the max
        moveDx = std::clamp(moveDx, -MAXMOVE, +MAXMOVE);
        moveDy = std::clamp(moveDy, -MAXMOVE, +MAXMOVE);

        // Try move in the y direction first
        if (!P_TryMove(mobj, mobj.x, mobj.y + moveDy)) {
            // Y move failed: try move in the x direction and kill all velocity in disallowed movement directions
            if (!P_TryMove(mobj, mobj.x + moveDx, mobj.y)) {
                mobj.momy = 0;
                mobj.momx = 0;
            } else {
                mobj.momx = moveDx;
                mobj.momy = 0;
            }
        } else {
            // Y move succeeded: kill all velocity in the x direction
            mobj.momx = 0;
            mobj.momy = moveDy;
        }
    }

    // Trigger line specials if we crossed a special line.
    // PsyDoom: we now allow multiple lines to be crossed at once, depending on game settings!
    #if PSYDOOM_MODS
    {
        // Get what special lines were crossed
        line_t** ppSpecialLines = nullptr;
        uint32_t numSpecialLines = 0;
        P_GetSpecialLines(ppSpecialLines, numSpecialLines);

        // Cross them in the same order that the code detected they were crossed in
        for (uint32_t lineIdx = 0; lineIdx < numSpecialLines; ++lineIdx) {
            ASSERT(ppSpecialLines[lineIdx]);
            line_t& line = *ppSpecialLines[lineIdx];
            P_CrossSpecialLine(line, mobj);
        }
    }
    #else
        line_t* const pSpecialLineToCross = gpSpecialLine;

        if (pSpecialLineToCross) {
            P_CrossSpecialLine(*pSpecialLineToCross, mobj);
        }
    #endif
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Do movement in the XY direction for the player map object and apply friction
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_PlayerXYMovement(mobj_t& mobj) noexcept {
    // Physically move the player and cross special lines, touch things etc.
    P_PlayerMove(mobj);

    // No friction when the player is in the air
    if (mobj.z > mobj.floorz)
        return;

    // If the player is a corpse and over a step then don't apply friction: slide down it
    if ((mobj.flags & MF_CORPSE) && (mobj.floorz != mobj.subsector->sector->floorheight))
        return;

    // If the player has reached a low enough velocity then stop completely, otherwise apply friction
    if ((std::abs(mobj.momx) < STOPSPEED) && (std::abs(mobj.momy) < STOPSPEED)) {
        mobj.momx = 0;
        mobj.momy = 0;
    } else {
        mobj.momx = d_rshift<8>(mobj.momx) * d_rshift<8>(FRICTION);
        mobj.momy = d_rshift<8>(mobj.momy) * d_rshift<8>(FRICTION);
    }
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Does height clipping and movement z velocity for the player thing.
// Also applies gravity when the thing is in the air.
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_PlayerZMovement(mobj_t& mobj) noexcept {
    player_t& player = *mobj.player;

    // Do smooth stepping up a step
    if (mobj.z < mobj.floorz) {
        player.viewheight -= mobj.floorz - mobj.z;                              // Adjust the view height by the difference
        player.deltaviewheight = d_rshift<2>(VIEWHEIGHT - player.viewheight);   // This does a nice curved motion up (fast at first, then slowing down)
    }

    // Advance z position by the velocity
    mobj.z += mobj.momz;

    // Clip z position against the floor and apply gravity
    if (mobj.z <= mobj.floorz) {
        // Hitting the floor: kill any downward velocity and clamp z position to the floor
        if (mobj.momz < 0) {
            if (mobj.momz < -GRAVITY * 2) {
                // If we hit the floor hard then play the 'oof' sound and do the hard floor impact (if allowed)
                player.deltaviewheight = d_rshift<3>(mobj.momz);
                S_StartSound(&mobj, sfx_oof);
            }

            mobj.momz = 0;
        }

        mobj.z = mobj.floorz;
    }
    else {
        // Not hitting the floor: do acceleration due to gravity.
        // Gravity is doubled on the first falling tic to account for the fact that it was always being applied, even on the floor:
        const fixed_t gravity = P_GetGravity();
        mobj.momz -= (mobj.momz == 0) ? gravity : gravity / 2;
    }

    // Clip height against the ceiling and kill upwards velocity if we're hitting it
    if (mobj.z + mobj.height > mobj.ceilingz) {
        mobj.z = mobj.ceilingz - mobj.height;
        mobj.momz = std::min(mobj.momz, 0);
    }
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Do movement and state transitions for the player's map object
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_PlayerMobjThink(mobj_t& mobj) noexcept {
    // Do XY and Z movement and velocity updates as required
    if ((mobj.momx != 0) || (mobj.momy != 0)) {
        P_PlayerXYMovement(mobj);
    }

    if ((mobj.z != mobj.floorz) || (mobj.momz != 0)) {
        P_PlayerZMovement(mobj);
    }

    // Transition to the next state if required, or never if the tic count is '-1' (forever).
    // Oddly, this logic does not call a state transition function for the player thing - not required because of player control?
    if (mobj.tics == -1)
        return;

    mobj.tics--;

    if (mobj.tics <= 0) {
        // Time to go to the next state: but don't call the state transition function as that's not needed for the player
        state_t& newState = gStates[mobj.state->nextstate];
        mobj.state = &newState;
        mobj.tics = newState.tics;
        mobj.sprite = newState.sprite;
        mobj.frame = newState.frame;
    }
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Reads inputs to decide how much the player will try to move and turn for a tic
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_BuildMove(player_t& player) noexcept {
    // Grab some useful stuff: elapsed vblanks, current and old inputs etc.
    const int32_t elapsedVBlanks = gPlayersElapsedVBlanks[gPlayerNum];
    const bool bFinalDoomMovementMode = Game::gSettings.bUseFinalDoomPlayerMovement;

    #if PSYDOOM_MODS
        const TickInputs& inputs = gTickInputs[gPlayerNum];
        const TickInputs& oldInputs = gOldTickInputs[gPlayerNum];

        const bool bTurnLeft = inputs.fTurnLeft();
        const bool bOldTurnLeft = oldInputs.fTurnLeft();
        const bool bTurnRight = inputs.fTurnRight();
        const bool bOldTurnRight = oldInputs.fTurnRight();
        const bool bStrafeLeft = inputs.fStrafeLeft();
        const bool bStrafeRight = inputs.fStrafeRight();
        const bool bMoveForward = inputs.fMoveForward();
        const bool bMoveBackward = inputs.fMoveBackward();
        const bool bRun = inputs.fRun();
        const bool bStrafe = inputs.fStrafe();
    #else
        const uint32_t curBtns = gTicButtons[gPlayerNum];
        const uint32_t oldBtns = gOldTicButtons[gPlayerNum];
        const padbuttons_t* const pBtnBindings = gpPlayerCtrlBindings[gPlayerNum];

        const bool bTurnLeft = (curBtns & PAD_LEFT);
        const bool bOldTurnLeft = (oldBtns & PAD_LEFT);
        const bool bTurnRight = (curBtns & PAD_RIGHT);
        const bool bOldTurnRight = (oldBtns & PAD_RIGHT);
        const bool bStrafeLeft = (curBtns & pBtnBindings[cbind_strafe_left]);
        const bool bStrafeRight = (curBtns & pBtnBindings[cbind_strafe_right]);
        const bool bMoveForward = (curBtns & PAD_UP);
        const bool bMoveBackward = (curBtns & PAD_DOWN);
        const bool bRun = (curBtns & pBtnBindings[cbind_run]);
        const bool bStrafe = (curBtns & pBtnBindings[cbind_strafe]);
    #endif

    // Do turn acceleration if turn is held continously for 2 frames or more
    const bool bLeftTurnAccel = (bTurnLeft && bOldTurnLeft);
    const bool bRightTurnAccel = (bTurnRight && bOldTurnRight);

    if (bLeftTurnAccel || bRightTurnAccel) {
        player.turnheld = std::min(player.turnheld + 1, TURN_ACCEL_TICS - 1);
    } else {
        player.turnheld = 0;
    }

    // Initially assume no turning or movement
    player.angleturn = 0;
    player.sidemove = 0;
    player.forwardmove = 0;

    // What movement speed is being used, run or normal?
    const uint32_t speedMode = (bRun) ? 1 : 0;

    // Do strafe left/right controls
    const bool bAllowMoveCancel = Game::gSettings.bAllowMovementCancellation;

    if (bFinalDoomMovementMode) {
        if (bStrafeLeft) {
            player.sidemove -= SIDE_MOVE_FDOOM[speedMode];
        }

        if (bStrafeRight && (bAllowMoveCancel || (!bStrafeLeft))) {
            player.sidemove += SIDE_MOVE_FDOOM[speedMode];
        }
    } else {
        // N.B: Applying the direction sign here before dividing is *VERY* important for demo compatibility
        if (bStrafeLeft) {
            player.sidemove += (-SIDE_MOVE_DOOM[speedMode] * elapsedVBlanks) / VBLANKS_PER_TIC;
        }

        if (bStrafeRight && (bAllowMoveCancel || (!bStrafeLeft))) {
            player.sidemove += (+SIDE_MOVE_DOOM[speedMode] * elapsedVBlanks) / VBLANKS_PER_TIC;
        }
    }

    // Do turning or strafing controls (if strafe button held)
    const bool bDoFastTurn = ((speedMode != 0) && (!bMoveForward) && (!bMoveBackward));
    const bool bAllowTurnCancel = Game::gSettings.bAllowTurningCancellation;

    const int32_t psxMouseSensitivityX = (gPsxMouseSensitivity * 100 * FRACUNIT) / 92;
    const int32_t psxMouseSensitivityY = 3000;
    const int32_t psxMouseTimeScale = (Game::gSettings.bUseFinalDoomPlayerMovement) ? 1 : elapsedVBlanks;
    const int32_t psxMouseMoveX = inputs.psxMouseDx * psxMouseSensitivityX * psxMouseTimeScale;
    const int32_t psxMouseMoveY = inputs.psxMouseDy * psxMouseSensitivityY * psxMouseTimeScale;

    if (bStrafe) {
        // Strafe button held: turn buttons and x psx mouse movements translate to strafing.
        if (bFinalDoomMovementMode) {
            if (bTurnLeft) {
                player.sidemove -= SIDE_MOVE_FDOOM[speedMode];
            }

            if (bTurnRight && (bAllowTurnCancel || (!bTurnLeft))) {
                player.sidemove += SIDE_MOVE_FDOOM[speedMode];
            }
        } else {
            // Note that for the original Doom this action also overwrites any strafe buttons held.
            // N.B: Applying the direction sign here before dividing is *VERY* important for demo compatibility.
            if (bTurnLeft) {
                player.sidemove = (-SIDE_MOVE_DOOM[speedMode] * elapsedVBlanks) / VBLANKS_PER_TIC;
            }

            if (bTurnRight && (bAllowTurnCancel || (!bTurnLeft))) {
                player.sidemove = (+SIDE_MOVE_DOOM[speedMode] * elapsedVBlanks) / VBLANKS_PER_TIC;
            }
        }

        // Handle Final Doom psx mouse movements that become strafing
        player.sidemove += psxMouseMoveX;
    }
    else {
        // No strafe button held: do normal turning.
        // 
        // PsyDoom: we now only do this if playing back a demo which restricts turning to 30 Hz ticks (classic, GEC demo formats).
        // In all other cases turning movements are now done outside of 30 Hz ticks and only committed here.
        #if PSYDOOM_MODS
            const bool bDoTurning = DemoPlayer::isPlayerTurning30HzCapped();
        #else
            const bool bDoTurning = true;
        #endif

        if (bDoTurning) {
            fixed_t turnAmt = 0;

            if (bDoFastTurn) {
                // Do fast turning when run is pressed and we are not moving forward/back
                const fixed_t turnSpeed = (bFinalDoomMovementMode) ? FAST_ANGLE_TURN_FDOOM[player.turnheld] : FAST_ANGLE_TURN_DOOM[player.turnheld];

                if (bTurnLeft) {
                    turnAmt += turnSpeed;
                }

                if (bTurnRight && (bAllowTurnCancel || (!bTurnLeft))) {
                    turnAmt -= turnSpeed;
                }
            } else {
                // Otherwise do the slow turning speed
                const fixed_t turnSpeed = (bFinalDoomMovementMode) ? ANGLE_TURN_FDOOM[player.turnheld] : ANGLE_TURN_DOOM[player.turnheld];

                if (bTurnLeft) {
                    turnAmt += turnSpeed;
                }

                if (bTurnRight && (bAllowTurnCancel || (!bTurnLeft))) {
                    turnAmt -= turnSpeed;
                }
            }

            // Apply the turn amount and also adjust for framerate (original Doom only)
            if (!bFinalDoomMovementMode) {
                turnAmt *= elapsedVBlanks;
                turnAmt /= VBLANKS_PER_TIC;
            }

            player.angleturn = (angle_t) d_lshift<TURN_TO_ANGLE_SHIFT>(turnAmt);

            // Apply Final Doom mouse turning also
            player.angleturn -= psxMouseMoveX;
        }
    }

    // Final Doom classic demo playback only: try and do a 'use' action when the mouse is double clicked quickly
    if (inputs.fPsxMouseUse()) {
        if ((player.psxMouseUseCountdown > 0) && (player.psxMouseUseCountdown < 10)) {
            // Rapid double click with the psx mouse done: try and do a use action and force the user to release the mouse before using again
            player.psxMouseUse = true;
            player.psxMouseUseCountdown = 0;
        } else if (player.psxMouseUseCountdown != 0) {
            // Initial click: you must release the mouse for at least 2 ticks then click again for the use action to register
            player.psxMouseUseCountdown = 11;
        }
    } else {
        // Mouse buttons not pressed: reduce the time the player has left to do the double click (or allow initial click again)
        player.psxMouseUseCountdown--;
    }

    // PsyDoom: apply analog turning movements; this has already been adjusted for framerate, so is applied directly.
    // We ignore the new turning system however when playing back classic demos.
    #if PSYDOOM_MODS
        if (DemoPlayer::getPlayingDemoFormat() != DemoFormat::Classic) {
            player.angleturn += inputs.getAnalogTurn();
        }
    #endif

    // Do forward/backward movement controls
    if (bFinalDoomMovementMode) {
        if (bMoveForward) {
            player.forwardmove += FORWARD_MOVE_FDOOM[speedMode];
        }

        if (bMoveBackward && (bAllowMoveCancel || (!bMoveForward))) {
            player.forwardmove -= FORWARD_MOVE_FDOOM[speedMode];
        }
    } else {
        // N.B: Applying the direction sign here before dividing is *VERY* important for demo compatibility
        if (bMoveForward) {
            player.forwardmove += (+FORWARD_MOVE_DOOM[speedMode] * elapsedVBlanks) / VBLANKS_PER_TIC;
        }

        if (bMoveBackward && (bAllowMoveCancel || (!bMoveForward))) {
            player.forwardmove += (-FORWARD_MOVE_DOOM[speedMode] * elapsedVBlanks) / VBLANKS_PER_TIC;
        }
    }

    // Apply Final Doom mouse movement also
    player.forwardmove -= psxMouseMoveY;

    // PsyDoom: do analog movements and also cancel any digital movement if using the analog controller
    #if PSYDOOM_MODS
        if (inputs.getAnalogForwardMove() != 0) {
            player.forwardmove = 0;
        }

        if (inputs.getAnalogSideMove() != 0) {
            player.sidemove = 0;
        }

        if (bFinalDoomMovementMode) {
            player.forwardmove += FixedMul(inputs.getAnalogForwardMove(), FORWARD_MOVE_FDOOM[speedMode]);
            player.sidemove += FixedMul(inputs.getAnalogSideMove(), SIDE_MOVE_FDOOM[speedMode]);
        } else {
            const fixed_t forwardMove = FixedMul(inputs.getAnalogForwardMove(), FORWARD_MOVE_DOOM[speedMode]);
            const fixed_t sideMove = FixedMul(inputs.getAnalogSideMove(), SIDE_MOVE_DOOM[speedMode]);

            player.forwardmove += (forwardMove * elapsedVBlanks) / VBLANKS_PER_TIC;
            player.sidemove += (sideMove * elapsedVBlanks) / VBLANKS_PER_TIC;
        }
    #endif

    // Clamp movement amounts: this was added in Final Doom
    const fixed_t maxForwardMove = (bFinalDoomMovementMode) ? FORWARD_MOVE_FDOOM[1] : (FORWARD_MOVE_DOOM[1] * elapsedVBlanks) / VBLANKS_PER_TIC;
    const fixed_t maxSideMove = (bFinalDoomMovementMode) ? SIDE_MOVE_FDOOM[1] : (SIDE_MOVE_DOOM[1] * elapsedVBlanks) / VBLANKS_PER_TIC;

    player.forwardmove = std::clamp(player.forwardmove, -maxForwardMove, +maxForwardMove);
    player.sidemove = std::clamp(player.sidemove, -maxSideMove, +maxSideMove);

    // PsyDoom: apply 'turbo mode' (2x speed) cheat if enabled
    #if PSYDOOM_MODS
        if (Game::gSettings.bTurboMode) {
            player.forwardmove *= 2;
            player.sidemove *= 2;
        }
    #endif

    // If the player is not moving at all change the animation frame to standing
    mobj_t& mobj = *player.mo;

    if ((mobj.momx == 0) && (mobj.momy == 0) && (player.forwardmove == 0) && (player.sidemove == 0)) {
        // Only switch to the standing frame if we are running; if we are already standing do nothing:
        state_t& state = *mobj.state;

        const bool bIsInRunState = (
            (&state == &gStates[S_PLAY_RUN1]) ||
            (&state == &gStates[S_PLAY_RUN2]) ||
            (&state == &gStates[S_PLAY_RUN3]) ||
            (&state == &gStates[S_PLAY_RUN4])
        );

        if (bIsInRunState) {
            P_SetMobjState(mobj, S_PLAY);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Applies the specified amount of velocity to the player's map object in the given direction
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_Thrust(player_t& player, const angle_t angle, const fixed_t amount) noexcept {
    mobj_t& mobj = *player.mo;
    const fixed_t timeScale = (Game::gSettings.bUseFinalDoomPlayerMovement) ? gPlayersElapsedVBlanks[gPlayerNum] : 1;
    const fixed_t scaledAmount = amount * timeScale;

    mobj.momx += d_rshift<8>(scaledAmount) * d_rshift<8>(gFineCosine[angle >> ANGLETOFINESHIFT]);
    mobj.momy += d_rshift<8>(scaledAmount) * d_rshift<8>(gFineSine[angle >> ANGLETOFINESHIFT]);
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Computes the player's current view height (for rendering) and does smooth stepping up stairs.
// Also applies view bobbing to the view height.
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_CalcHeight(player_t& player) noexcept {
    // Compute movement bobbing amount
    mobj_t& mobj = *player.mo;

    {
        int32_t speedX = mobj.momx;
        int32_t speedY = mobj.momy;

        #if PSYDOOM_MODS
        {
            // PsyDoom: if we are doing 30 Hz ticks then adjust the speed fed into the view bobbing calculation to account for doubled friction @ 30 Hz.
            // Undo 1 tick of friction to get a bobbing amount closer to what would happen if the game ran at 15 Hz.
            const uint32_t worldTickVblanks = (Game::gSettings.bUsePalTimings) ? 3 : 4;

            if (Game::gSettings.bFixViewBobStrength && (gElapsedVBlanks < worldTickVblanks)) {
                const fixed_t oldSpeedX = speedX;
                const fixed_t oldSpeedY = speedY;
                speedX = FixedDiv(speedX, FRICTION);
                speedY = FixedDiv(speedY, FRICTION);

                // If we are midway between 30 Hz and 15 Hz then lerp between the old and new (adjusted) speed values
                if (gElapsedVBlanks == 3) {
                    speedX = (speedX + oldSpeedX) / 2;
                    speedY = (speedY + oldSpeedY) / 2;
                }
            }

            // PsyDoom: if we are applying the movement input latency tweak then un-apply this frame's friction for the bob calculation.
            // This makes the bob operate at a similar strength to when the latency tweak is disabled.
            // We need this adjustment because prior to the latency tweak view bobbing was calculated BEFORE friction was applied.
            if (Game::gSettings.bUseMoveInputLatencyTweak) {
                speedX = FixedDiv(speedX, FRICTION);
                speedY = FixedDiv(speedY, FRICTION);
            }
        }
        #endif

        speedX = d_rshift<8>(speedX);
        speedY = d_rshift<8>(speedY);
        player.bob = (speedX * speedX) + (speedY * speedY);

        // For some reason the PAL version divides this amount by 3 rather than 16...
        // PsyDoom: fix that and make the behavior consistent for both versions if the view bobbing fix is enabled.
        if (Game::gSettings.bUsePalTimings && (!Game::gSettings.bFixViewBobStrength)) {
            player.bob /= 3;
        } else {
            player.bob = d_rshift<4>(player.bob);
        }

        // PsyDoom: scale the strength of view bobbing by the user preferences
        #if PSYDOOM_MODS
            player.bob = FixedMul(player.bob, Game::gSettings.viewBobbingStrengthFixed);
        #endif

        player.bob = std::min(player.bob, MAXBOB);
    }

    // When we are not on the ground just set the view z based on map object z and clamp below the ceiling
    const fixed_t maxViewZ = mobj.ceilingz - 4 * FRACUNIT;

    if (!gbOnGround) {
        player.viewz = mobj.z + VIEWHEIGHT;
        player.viewz = std::min(player.viewz, maxViewZ);    // Don't get too close to the ceiling!
        return;
    }

    // Do view height movements due smooth stepping up stairs
    if (player.playerstate == PST_LIVE) {
        player.viewheight += player.deltaviewheight;

        // Raised too far?
        if (player.viewheight > VIEWHEIGHT) {
            player.viewheight = VIEWHEIGHT;
            player.deltaviewheight = 0;     // Stop moving up
        }

        // Gone down to low?
        if (player.viewheight < VIEWHEIGHT / 2) {
            player.viewheight = VIEWHEIGHT / 2;
            player.deltaviewheight = std::max(player.deltaviewheight, 1);   // Start moving up
        }

        // If we are moving up still, accelerate the the movement over time
        if (player.deltaviewheight != 0) {
            player.deltaviewheight += FRACUNIT / 2;

            // Don't stop moving yet if we were moving down previously but now have flipped around
            if (player.deltaviewheight == 0) {
                player.deltaviewheight = 1;
            }
        }
    }

    // Compute the bob amplitude, which uses a sine wave as it's basis.
    // We step a certain amount of the bob with every vblank elapsed:
    constexpr int32_t BOB_PHASE_TIME = (VBLANKS_PER_SEC * 2) / 3;
    constexpr int32_t BOB_PHASE_STEP = FINEANGLES / BOB_PHASE_TIME;

    const uint32_t bobPhase = (gTicCon * BOB_PHASE_STEP) & FINEMASK;
    const fixed_t bobAmplitude = gFineSine[bobPhase];

    // Compute the final view z based on map object z, view height and bob amount
    const fixed_t oldViewZ = player.viewz;
    player.viewz = mobj.z + player.viewheight + d_rshift<17>(player.bob) * bobAmplitude;

    // Clamp the view z so it's not to close to the ceiling (if need be)
    const bool bCeilingClipViewZ = (player.viewz > maxViewZ);

    if (bCeilingClipViewZ) {
        player.viewz = maxViewZ;
    }

    // PsyDoom: if this player's view is sticking into the ceiling then that needs to be handled differently for interpolation.
    // Assume this is due to a crusher operating, which ticks at 15 Hz (world/mobj tick rate) instead of the player's 30 Hz.
    // Because the crusher ticks at a slower rate, we need to interpolate this motion over a longer period.
    // Make a note of how much the player's view was pushed by the crusher moving here:
    #if PSYDOOM_MODS
        const bool bIsCurrentPlayer = (&player == &gPlayers[gCurPlayerIndex]);

        // Is the current player's view being clipped by a crushing ceiling?
        if (bIsCurrentPlayer && bCeilingClipViewZ) {
            // Only account for this difference on the even frames where we do a world/mobj simulation too.
            // On the odd player (30 Hz) frames we've already noted this amount so don't need to add it again.
            if (!gbOldViewZIsPushed) {
                gViewPushedZ += player.viewz - oldViewZ;
            }
        }
    #endif
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Applies turn rotation and velocity due to movement inputs to the player.
// Also puts the player into the run animation if moving.
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_MovePlayer(player_t& player) noexcept {
    // Apply angle turning
    mobj_t& mobj = *player.mo;

    if (Game::gSettings.bUseFinalDoomPlayerMovement) {
        // PsyDoom: we only need to apply the vblank scale if playing back a demo which restricts turning to be within 30 Hz ticks
        #if PSYDOOM_MODS
            if (DemoPlayer::isPlayerTurning30HzCapped()) {
                mobj.angle += player.angleturn * gPlayersElapsedVBlanks[gPlayerNum];
            } else {
                mobj.angle += player.angleturn;
            }
        #else
            mobj.angle += player.angleturn * gPlayersElapsedVBlanks[gPlayerNum];
        #endif
    } else {
        mobj.angle += player.angleturn;
    }

    // Save whether we are on the ground
    gbOnGround = (mobj.z <= mobj.floorz);

    // Apply side and forward/backward movement velocity
    if ((player.forwardmove != 0) && gbOnGround) {
        P_Thrust(player, mobj.angle, player.forwardmove);
    }

    if ((player.sidemove != 0) && gbOnGround) {
        P_Thrust(player, mobj.angle - ANG90, player.sidemove);
    }

    // Switch to the running animation frames if moving and currently showing the standing still ones
    const bool bIsMoving = ((player.forwardmove != 0) || (player.sidemove != 0));

    if (bIsMoving && (mobj.state == &gStates[S_PLAY])) {
        P_SetMobjState(mobj, S_PLAY_RUN1);
    }
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Does updates for the player when dead: drops down the view height, weapons and faces the killer etc.
//------------------------------------------------------------------------------------------------------------------------------------------
static void P_DeathThink(player_t& player) noexcept {
    // Drop the currently held weapon
    P_MovePsprites(player);

    // Lower the view to the ground
    if (player.viewheight > 8 * FRACUNIT) {
        player.viewheight -= FRACUNIT;
    }

    // Is the player on the ground now?
    mobj_t& playerMobj = *player.mo;
    gbOnGround = (playerMobj.z <= playerMobj.floorz);

    // Update the current view height
    P_CalcHeight(player);

    // If we were killed by an enemy then turn to face it
    mobj_t* const pAttacker = player.attacker;

    if (pAttacker && (pAttacker != &playerMobj)) {
        // Killed by an enemy: turn to face it, then fade out the reddamage tint
        const angle_t angleToAttacker = R_PointToAngle2(playerMobj.x, playerMobj.y, pAttacker->x, pAttacker->y);
        const angle_t angleDelta = angleToAttacker - playerMobj.angle;

        if ((angleDelta >= ANG5) && (angleDelta <= (angle_t) -ANG5)) {
            // Still turning to face the killer: turn the right way
            if (angleDelta < ANG180) {
                playerMobj.angle += ANG5;
            } else {
                playerMobj.angle -= ANG5;
            }
        }
        else {
            // Now looking directly at the killer: snap to it's angle and fade out the damage tint
            playerMobj.angle = angleToAttacker;

            if (player.damagecount != 0) {
                player.damagecount--;
            }
        }
    } 
    else {
        // Not killed by an enemy: just fade out the red damage tint
        if (player.damagecount != 0) {
            player.damagecount--;
        }
    }

    // Respawn if the right buttons are pressed and the player's view has dropped enough
    #if PSYDOOM_MODS
        const bool bRespawnBtnPressed = gTickInputs[gPlayerNum].fRespawn();
    #else
        const bool bRespawnBtnPressed = (gTicButtons[gPlayerNum] & (PAD_ACTION_BTNS | PAD_SHOULDER_BTNS));
    #endif

    if (bRespawnBtnPressed && (player.viewheight <= 8 * FRACUNIT)) {
        player.playerstate = PST_REBORN;
    }
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Does all player updates such as movement, controls, weapon switching and so forth.
// This is called 30 times a second for the NTSC version.
//------------------------------------------------------------------------------------------------------------------------------------------
void P_PlayerThink(player_t& player) noexcept {
    // PsyDoom: don't tick the player if the external camera is active, but kill all momentum
    #if PSYDOOM_MODS
        if (gExtCameraTicsLeft > 0) {
            mobj_t& mobj = *player.mo;
            mobj.momx = 0;
            mobj.momy = 0;
            mobj.momz = 0;
            return;
        }
    #endif

    // Grab the current and previous inputs
    #if PSYDOOM_MODS
        const TickInputs& inputs = gTickInputs[gPlayerNum];
        const TickInputs& oldInputs = gOldTickInputs[gPlayerNum];

        const bool bPrevWeapon = (inputs.fPrevWeapon() && (!oldInputs.fPrevWeapon()));
        const bool bNextWeapon = (inputs.fNextWeapon() && (!oldInputs.fNextWeapon()));
        const bool bUse = inputs.fUse();
        const bool bAttack = inputs.fAttack();
    #else
        const uint32_t curBtns = gTicButtons[gPlayerNum];
        const uint32_t oldBtns = gOldTicButtons[gPlayerNum];
        const padbuttons_t* pBtnBindings = gpPlayerCtrlBindings[gPlayerNum];

        const bool bPrevWeapon = Utils::padBtnJustPressed(pBtnBindings[cbind_prev_weapon], curBtns, oldBtns);
        const bool bNextWeapon = Utils::padBtnJustPressed(pBtnBindings[cbind_next_weapon], curBtns, oldBtns);
        const bool bUse = (curBtns & pBtnBindings[cbind_use]);
        const bool bAttack = (curBtns & pBtnBindings[cbind_attack]);
    #endif

    // Do weapon switching if the player is still alive (and even if paused)
    if (player.playerstate == PST_LIVE) {
        // Get the current weapon being switched to, or the ready weapon if not switching weapons
        const weapontype_t curWeaponType = (player.pendingweapon == wp_nochange) ? player.readyweapon : player.pendingweapon;

        // Get the current micro box that is highlighted (for this weapon)
        int32_t weaponMicroBoxIdx = WEAPON_MICRO_INDEXES[curWeaponType];
        int32_t nextWeaponIdx = weaponMicroBoxIdx;

        // See if we are switching weapons
        if (bPrevWeapon) {
            // Try to go to the previous weapon
            if ((weaponMicroBoxIdx == 0) && player.weaponowned[wp_chainsaw]) {
                // When we are on the 1st slot and own the chainsaw then allow toggling between fists and chainsaw.
                // This assignment doesn't mean anything, it's just set to trigger a weapon change (see below).
                weaponMicroBoxIdx = 1;
            }
            else if (weaponMicroBoxIdx > 0) {
                // If above the first slot go to the previous weapon.
                // Keep going back until we find a weapon that is owned - which we should always have (fists/pistol).
                nextWeaponIdx--;

                while ((!player.weaponowned[nextWeaponIdx]) && (nextWeaponIdx > 0)) {
                    nextWeaponIdx--;
                }
            }
        }
        else if (bNextWeapon) {
            // Go to the next weapon: keep incrementing the weapon index until we find one that is owned or hit the last microbox index
            if (weaponMicroBoxIdx < 8) {
                nextWeaponIdx++;

                while ((!player.weaponowned[nextWeaponIdx]) && (nextWeaponIdx < 8)) {
                    nextWeaponIdx++;
                }
            }

            // If we incremented onto the chainsaw (value 8) then go back to the bfg (wraparound is not allowed)
            if (nextWeaponIdx == wp_chainsaw) {
                nextWeaponIdx = wp_bfg;

                // If the bfg is not owned keep going back in the weapon list until we find one that is owned (there will always be something owned)
                while (!player.weaponowned[nextWeaponIdx]) {
                    nextWeaponIdx--;
                }
            }
        }

        // Did we decide to change weapon?
        if (nextWeaponIdx != weaponMicroBoxIdx) {
            // If we are on the fists and triggering a change then toggle between those and the chainsaw
            if ((nextWeaponIdx == wp_fist) && (player.readyweapon != wp_chainsaw)) {
                nextWeaponIdx = (player.weaponowned[wp_chainsaw] != 0) ? wp_chainsaw : wp_fist;
            }

            // Actually change the weapon
            player.pendingweapon = (weapontype_t) nextWeaponIdx;
        }

        // PsyDoom: do direct weapon switching too if requested, if the weapon is valid and different to the current equipped weapon
        #if PSYDOOM_MODS
            if ((inputs.directSwitchToWeapon != wp_nochange) && (inputs.directSwitchToWeapon < NUMWEAPONS)) {
                const bool bAllowSwitch = (
                    player.weaponowned[inputs.directSwitchToWeapon] && (            // Must own the weapon to switch to it
                        (inputs.directSwitchToWeapon != player.readyweapon) ||      // Don't allow switching to already held weapon 
                        (player.pendingweapon != wp_nochange)                       // But DO allow switching back to the held back if another switch was started (cancel switch)
                    )
                );

                if (bAllowSwitch) {
                    player.pendingweapon = (weapontype_t) inputs.directSwitchToWeapon;
                }
            }
        #endif
    }

    // PsyDoom: check for quick save and load keys in singleplayer
    #if PSYDOOM_MODS
        if (gNetGame == gt_single) {
            if (inputs.fQuicksave() && (!oldInputs.fQuicksave())) {
                // Only allow quicksave if the player is not dead, otherwise pointless.
                // You can still manually save when dead however, if you really really want to...
                if ((player.health > 0) && (player.playerstate == PST_LIVE)) {
                    gbDoQuicksave = true;
                }
            }

            if (inputs.fQuickload() && (!oldInputs.fQuickload())) {
                gbDoQuickload = true;
            }
        }
    #endif

    // Updates for when the game is NOT paused
    if (!gbGamePaused) {
        // Do physical movements due to velocity and state transitions for the player.
        // PsyDoom: do this AFTER gathering inputs to reduce input lag, except in the case of demos (for compatibility).
        #if PSYDOOM_MODS
            const bool bUseOrigMovement = (!Game::gSettings.bUseMoveInputLatencyTweak);
        #else
            const bool bUseOrigMovement = false;
        #endif

        mobj_t& playerMobj = *player.mo;

        // Originally the player was physically moved BEFORE gathering inputs
        if (bUseOrigMovement) {
            P_PlayerMobjThink(playerMobj);
        }

        // PsyDoom: update any 'Voodoo dolls' that exist for this player.
        // This prevents the dolls from having their animation stuck and also allows them to respond to forces.
        // Sector specials like slime are NOT supported on the doll however and do not have any effect - this is just a barebones update.
        #if PSYDOOM_MODS
            for (mobj_t* pMobj = gMobjHead.next; pMobj != &gMobjHead; pMobj = pMobj->next) {
                // Is it a player thing belonging to this player?
                if ((pMobj->type == MT_PLAYER) && (pMobj->player == &player)) {
                    // Is it not the thing the player is controlling, and therefore a 'Voodoo doll'?
                    if (pMobj != &playerMobj) {
                        P_PlayerMobjThink(*pMobj);
                    }
                }
            }
        #endif

        // Gather inputs for the next move
        P_BuildMove(player);

        // Is the player dead? If so do death updates, otherwise do all the normal updates
        if (player.playerstate == PST_DEAD) {
            P_DeathThink(player);

            // PsyDoom: just being consistent with the 'alive' case if we are using tweaked movement
            if (!bUseOrigMovement) {
                P_PlayerMobjThink(playerMobj);
            }
        }
        else {
            // Attacking with the chainsaw causes some movement forward
            if (playerMobj.flags & MF_JUSTATTACKED) {
                player.angleturn = 0;
                player.forwardmove = FRACUNIT * 25 / 32;
                player.sidemove = 0;
                playerMobj.flags &= ~MF_JUSTATTACKED;
            }

            // Apply thrust unless delayed after exiting a teleporter etc.
            if (playerMobj.reactiontime == 0) {
                P_MovePlayer(player);
            } else {
                playerMobj.reactiontime--;
            }

            // PsyDoom: actually move the player at this point, AFTER inputs
            if (!bUseOrigMovement) {
                P_PlayerMobjThink(playerMobj);
            }

            // Adjust view height
            P_CalcHeight(player);

            // Do sector specials: lava and slime etc. hurting the player
            if (playerMobj.subsector->sector->special) {
                P_PlayerInSpecialSector(player);

                if (player.playerstate == PST_DEAD)
                    return;
            }

            // Use special lines if the use button is pressed
            if (bUse || player.psxMouseUse) {
                player.psxMouseUse = false;

                if (!player.usedown) {
                    P_UseLines(player);
                    player.usedown = true;
                }
            } else {
                player.usedown = false;
            }

            // Go into the attack state and update the status bar for this player if fire is pressed for a long time on certain weapons
            if (bAttack) {
                P_SetMobjState(playerMobj, S_PLAY_ATK1);
                player.attackdown++;

                // Should we do the grimmace face after fire has been pressed a long time?
                const bool bDoGrimmaceFace = (
                    (gPlayerNum == gCurPlayerIndex) &&
                    (player.attackdown > TICRATE * 2) &&
                    ((player.readyweapon == wp_chaingun) || (player.readyweapon == wp_plasma))
                );

                if (bDoGrimmaceFace) {
                    gStatusBar.specialFace = f_mowdown;
                }
            } else {
                // Not attacking this tic: reset the count for number of tics attacking
                player.attackdown = 0;
            }

            // Update the player's weapon sprites
            P_MovePsprites(player);

            // Tic special powers if enough time has elapsed and reduce their duration.
            if (gGameTic > gPrevGameTic) {
                // Note: berserk is special and counts up to diminish the red hue effect after picking it up.
                // It never runs out (theoretically it CAN if there is an overflow) unlike other powers...
                if (player.powers[pw_strength]) {
                    player.powers[pw_strength]++;
                }

                if (player.powers[pw_invulnerability] != 0) {
                    player.powers[pw_invulnerability]--;
                }

                if (player.powers[pw_invisibility]) {
                    player.powers[pw_invisibility]--;

                    // Remove the invisibility blend mask if it's out
                    if (player.powers[pw_invisibility] == 0) {
                        playerMobj.flags &= ~MF_BLEND_ADD_25;
                    } else {
                        // If we are getting low on invisibility (4 seconds or less) then blink it out:
                        if (player.powers[pw_invisibility] <= TICRATE * 4) {
                            if ((player.powers[pw_invisibility] & 7) == 0) {
                                playerMobj.flags ^= MF_BLEND_ADD_25;
                            }
                        }
                    }
                }

                if (player.powers[pw_infrared]) {
                    player.powers[pw_infrared]--;
                }

                if (player.powers[pw_ironfeet]) {
                    player.powers[pw_ironfeet]--;
                }

                // Reduce damage and bonus tints
                if (player.damagecount != 0) {
                    player.damagecount--;
                }

                if (player.bonuscount != 0) {
                    player.bonuscount--;
                }
            }
        }
    }
}

#if PSYDOOM_MODS
//------------------------------------------------------------------------------------------------------------------------------------------
// PsyDoom: initialize the new (framerate uncapped) turning system for the current player
//------------------------------------------------------------------------------------------------------------------------------------------
void P_PlayerInitTurning() noexcept {
    gLastPlayerTurnTime = std::chrono::high_resolution_clock::now();
    gPlayerUncommittedTurning = 0;
    gPlayerNextTickViewAngle = gPlayers[gCurPlayerIndex].mo->angle;
}

//------------------------------------------------------------------------------------------------------------------------------------------
// PsyDoom: do the newly added framerate uncapped turning for the current player.
// We now allow turning at any point in time and without any interpolation, to reduce input lag.
//------------------------------------------------------------------------------------------------------------------------------------------
void P_PlayerDoTurning() noexcept {
    const time_point_t now = std::chrono::high_resolution_clock::now();
    const player_t& player = gPlayers[gCurPlayerIndex];
    const bool bFinalDoomMovementMode = Game::gSettings.bUseFinalDoomPlayerMovement;

    // Only do these turning updates if the player is not dead, the game is active, and if we're not doing a demo.
    //
    // IMPORTANT: I previously had a call to 'Input::update()' here to get the very latest inputs but that caused bugs
    // and it should NOT be added back in. If inputs are updated here then any new events received might be consumed prior
    // to rendering, because we consume used input events once the game tick is processed. The rule of thumb is that no
    // event polling should be done WHILE game logic is being processed. Inputs should only be updated BEFORE the tick starts.
    //
    const bool bCanTurn = (
        (!gbDemoPlayback) &&
        (player.playerstate == PST_LIVE) &&
        (player.mo && (player.mo->reactiontime <= 0)) &&    // Disallow turning for a little bit after teleporting (original movement code did this too)
        (!gbGamePaused)
    );

    if (bCanTurn) {
        // Get how much time has elapsed in terms of 60 Hz ticks (NTSC vblanks).
        // Note that I'm deliberately NOT adjusting turn speed for PAL mode here, since turning is now independent of framerate anyway.
        // The 60 Hz reference point was just to scale the turn amount and treating PAL/NTSC the same keeps the turn speed consistent in both modes.
        const float timeDeltaF = std::chrono::duration<float>(now - gLastPlayerTurnTime).count();
        const float ticks60F = std::min(timeDeltaF * 60.0f, 60.0f);
        const fixed_t ticks60 = (fixed_t)(ticks60F * FRACUNIT);

        // Do keyboard turning if the 'strafe' button is not pressed (that makes turn keys act as strafe instead)
        const TickInputs& inputs = gTickInputs[gCurPlayerIndex];

        if (!inputs.fStrafe()) {
            const bool bAllowTurnCancel = Game::gSettings.bAllowTurningCancellation;
            fixed_t turnAmt = 0;

            if (inputs.fRun() && (!inputs.fMoveForward()) && (!inputs.fMoveBackward())) {
                // Do fast turning when run is pressed and we are not moving forward/back
                const fixed_t turnSpeed = (bFinalDoomMovementMode) ? FAST_ANGLE_TURN_FDOOM[player.turnheld] : FAST_ANGLE_TURN_DOOM[player.turnheld];

                if (inputs.fTurnLeft()) {
                    turnAmt += turnSpeed;
                }

                if (inputs.fTurnRight() && (bAllowTurnCancel || (!inputs.fTurnLeft()))) {
                    turnAmt -= turnSpeed;
                }
            } else {
                // Otherwise do the slow turning speed
                const fixed_t turnSpeed = (bFinalDoomMovementMode) ? ANGLE_TURN_FDOOM[player.turnheld] : ANGLE_TURN_DOOM[player.turnheld];

                if (inputs.fTurnLeft()) {
                    turnAmt += turnSpeed;
                }

                if (inputs.fTurnRight() && (bAllowTurnCancel || (!inputs.fTurnLeft()))) {
                    turnAmt -= turnSpeed;
                }
            }

            turnAmt = FixedMul((fixed_t)((float) turnAmt * PlayerPrefs::getTurnSpeedMultiplier()), ticks60);

            if (!bFinalDoomMovementMode) {
                turnAmt /= VBLANKS_PER_TIC;
            }

            gPlayerUncommittedTurning += (angle_t) d_lshift<TURN_TO_ANGLE_SHIFT>(turnAmt);
        }

        // Do analog controller turning from the gamepad
        {
            // Note: this axis value will be properly deadzone adjusted so that '0' starts at the end of the deadzone region
            const float axis = (
                Controls::getFloat(Controls::Binding::Analog_TurnRight) -
                Controls::getFloat(Controls::Binding::Analog_TurnLeft)
            );

            // Figure out how much of the high and low turn speeds to use; use the higher turn speed as the stick is pressed more.
            // Note: again I'm not adjusting for PAL mode since turning is independent of framerate, maintain the same turn speed in both PAL and NSTC mode.
            const float turnSpeedLow = (inputs.fRun()) ? Config::gGamepadFastTurnSpeed_Low : Config::gGamepadTurnSpeed_Low;
            const float turnSpeedHigh = (inputs.fRun()) ? Config::gGamepadFastTurnSpeed_High : Config::gGamepadTurnSpeed_High;
            const float turnSpeedMix = std::abs(axis);
            const float turnSpeed = (turnSpeedLow * (1.0f - turnSpeedMix) + turnSpeedHigh * turnSpeedMix) * PlayerPrefs::getTurnSpeedMultiplier();

            fixed_t turnAmt = FixedMul((fixed_t)(turnSpeed * axis), ticks60);
            turnAmt /= VBLANKS_PER_TIC;
            gPlayerUncommittedTurning -= (angle_t) d_lshift<TURN_TO_ANGLE_SHIFT>(turnAmt);
        }

        // Do turning from the mouse and consume the movements after we have counted them
        {
            const float axis = -Input::getMouseXMovement();
            const float turnSpeed = Config::gMouseTurnSpeed * PlayerPrefs::getTurnSpeedMultiplier();
            const fixed_t turnAmt = (fixed_t)(turnSpeed * axis);

            gPlayerUncommittedTurning += (angle_t) d_lshift<TURN_TO_ANGLE_SHIFT>(turnAmt);
            Input::consumeMouseMovements();
        }
    }

    // Remember the last time we did turning updates
    gLastPlayerTurnTime = now;
}

//------------------------------------------------------------------------------------------------------------------------------------------
// Called upon pausing the game; moves upcoming analog turning tick inputs back to being 'uncommitted' inputs.
// Required to avoid the view snapping back when pausing while turning, because tick inputs are essentially discarded upon pause.
//------------------------------------------------------------------------------------------------------------------------------------------
void P_UncommitTurningTickInputs() noexcept {
    TickInputs& tickInputs = gTickInputs[gCurPlayerIndex];

    // Note: for networked games we need to consider inputs for the next frame too
    if (gNetGame == gt_single) {
        gPlayerUncommittedTurning += tickInputs.getAnalogTurn();
        tickInputs.setAnalogTurn(0);
    } else {
        TickInputs& nextTickInputs = gNextTickInputs;
        gPlayerUncommittedTurning += tickInputs.getAnalogTurn();
        gPlayerUncommittedTurning += nextTickInputs.getAnalogTurn();
        gPlayerNextTickViewAngle -= tickInputs.getAnalogTurn();         // Don't double count, remove the turning from this since it's already in there
        gPlayerNextTickViewAngle -= nextTickInputs.getAnalogTurn();

        // Note: even though it's REALLY bad practice to modify the inputs we said we'd do in the next frame, all peers understand that when
        // the game is paused tick inputs are essentially discarded. So wiping the inputs in this way is ok to do in this limited situation:
        tickInputs.setAnalogTurn(0);
        nextTickInputs.setAnalogTurn(0);
    }
}
#endif  // #if PSYDOOM_MODS
